/*!
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

import React, { useContext, useRef } from 'react';

// eslint-disable-next-line max-len
const ContainerRefContext = React.createContext<React.RefObject<HTMLDivElement> | undefined>(undefined);

interface Props {
  children: React.ReactNode;
}

// containerRef is necessary to render for tooltips, modals, and dialogs
// This provider allows the containerRef to be accessed by any react component
export const ContainerRefProvider = ({ children }: Props) => {
  const containerRef = useRef<HTMLDivElement>(null);

  return (
    <ContainerRefContext.Provider value={containerRef}>
      <div ref={containerRef}>
        {children}
      </div>
    </ContainerRefContext.Provider>
  );
};

export const useContainerRef = () => useContext(ContainerRefContext);
