# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os

from airflow.models import DagBag
from airflow.settings import DAGS_FOLDER


def init_dagbag(app):
    """
    Create global DagBag for webserver and API. To access it use
    ``flask.current_app.dag_bag``.
    """
    if os.environ.get('SKIP_DAGS_PARSING') == 'True':
        app.dag_bag = DagBag(os.devnull, include_examples=False)
    else:
        app.dag_bag = DagBag(DAGS_FOLDER, read_dags_from_db=True)
